#include "des.h"

/**
 Simple DES
 이 알고리즘은 10 bit key를 사용하고, 데이터를 한 번에 8 bit씩 암호화 합니다.

 DEBUG 모드를 사용해서 모든 과정을 출력시킬 수 있습니다.
 만약 DEBUG 모드를 설정하지 않으면, 간단한 결과만을 출력시킵니다.
 DEBUG 모드는 기본적으로 꺼져있고, GCC에서는 -DDEBUG 옵션과 함께 컴파일을 하면 사용할 수 있습니다.
 Visual Studio 환경에서는 매크로를 사용하면 됩니다.

 이 코드는 GCC와 Visual Studio 환경에서 올바르게 동작합니다.

 코드 작성: John W. Clark
 수정 및 번역: CY. Kim
*/

typedef unsigned short uval; // 16 bit 부호 없는 수
// 8 bit와 10 bit 수는 그냥 uval 변수를 사용합니다.
// 이 것은 편법에 지나지 않습니다.

// s-Box를 생성합니다.
static const uval S0[4][4] = {
 { 1, 0, 3, 2 },
 { 3, 2, 1, 0 },
 { 0, 2, 1, 3 },
 { 3, 1, 3, 2 }
};
static const uval S1[4][4] = {
 { 0, 1, 2, 3 },
 { 2, 0, 1, 3 },
 { 3, 0, 1, 0 },
 { 2, 1, 0, 3 }
};

/*
 나는 아래 두 가지 이유로 매크로를 사용하였습니다.
 1. 함수 호출 시의 불필요한 오버헤드를 피할 수 있습니다.
 2. 정보보안을 전공하는 학생들에게 매우 많은 흥미를 유발할 수 있습니다.
 만약, GCC를 사용한다면 매크로 확장 방법을 다음 명령어로 확인할 수 있습니다. : gcc -E
*/

// Visual Studio 환경에서 디버깅 모드를 사용하려면, DEBUG 매크로를 정의합니다.
// #define DEBUG
#define DEBUG

// y를 x bit 만큼 쉬프팅 하여 1 또는 0인지 계산합니다. ( 참/거짓 )
#define B(x,y) ( ((y) >> (x)) & 0x01)

/*
 순열에 의하여 bit들을 재배치 시키고 변형합니다.
 이 것은 모두 매크로에 의해 정의되어 있습니다.
*/

// p10 9876543210 -> 7583609124
#define p10(x) ( B(7,x) << 9 | B(5,x) << 8 | B(8,x) << 7 | B(3,x) << 6 | \
  B(6,x) << 5 | B(0,x) << 4 | B(9,x) << 3 | B(1,x) << 2 | \
  B(2,x) << 1 | B(4,x) )

//p8 9876543210 -> 47362501
#define p8(x) ( B(4,x) << 7 | B(7,x) << 6 | B(3,x) << 5 | B(6,x) << 4 | \
  B(2,x) << 3 | B(5,x) << 2 | B(0,x) << 1 | B(1,x) )

//p4 3210 -> 2013
#define p4(x) ( B(2,x) << 3 | B(0,x) << 2 | B(1,x) << 1 | B(3,x) )

//ip 76543210 -> 62574031
#define ip(x) ( B(6,x) << 7 | B(2,x) << 6 | B(5,x) << 5 | B(7,x) << 4 | \
  B(4,x) << 3 | B(0,x) << 2 | B(3,x) << 1 | B(1,x) )

//ip_inv 76543210 -> 47531602
#define ip_inv(x) ( B(4,x) << 7 | B(7,x) << 6 | B(5,x) << 5 | B(3,x) << 4 | \
  B(1,x) << 3 | B(6,x) << 2 | B(0,x) << 1 | B(2,x) )

//ep 3210 -> 03212103
#define ep(x) ( B(0,x) << 7 | B(3,x) << 6 | B(2,x) << 5 | B(1,x) << 4 | \
  B(2,x) << 3 | B(1,x) << 2 | B(0,x) << 1 | B(3,x) )

//sw 76543210 -> 32107654
#define sw(x) ( B(3,x) << 7 | B(2,x) << 6 | B(1,x) << 5 | B(0,x) << 4 | \
  B(7,x) << 3 | B(6,x) << 2 | B(5,x) << 1 | B(4,x) )

// 실제로 아래의 코드를 사용했는데,
// x를 y bit 만큼 좌측으로 쉬프팅 하면, 전체 bit 수는 5 입니다.
#define ROT_L5(x,y) ((( (x) << (y) ) | ( (x) >> (5-y) )) & 0x1F )
// x의 오른쪽 5 bit를 가져옵니다.
#define RIGHT5(x) ( (x) & 0x1F )
// x의 왼쪽 5 bit를 가져옵니다.
#define LEFT5(x) ( ( (x) >> 5 ) & 0x1F )
// x의 오른쪽 4 bit를 가져옵니다.
#define RIGHT4(x) ( (x) & 0x0F )
// x의 왼쪽 4 bit를 가져옵니다.
#define LEFT4(x) ( ( (x) >> 4 ) & 0x0F )

// x를 10개의 bit 값으로 출력합니다.
void p_bits( uval x ) {
 char arr[13] = "00 0000 0000";
 int i, j;
 for( i=9, j=0; i>=0; --i, ++j ) {
  // 가독성 향상을 위하여 2번째 칸과 7번째 칸에 공백을 그대로 출력합니다.
  if( ( j == 2 ) || ( j == 7 ) ) {
   ++j;
  }
  if( B(i,x ) ) {
   arr[ j ] = '1';
  }
 }
 //printf( "%s\n", arr );
}

#ifdef DEBUG
// 불필요시 아래 디버깅 정보를 제거하세요.
void debugf( char * s, uval x ) {
 //printf( "%s", s );
 p_bits( x );
}
#else
// 비어있는 상태 정보를 생성합니다.
#define debugf(s,x);
#endif

// 실제 DES subkey를 구하는 함수입니다.
void subkeys( uval k, uval *k1, uval *k2 ) {
 //printf( "\n[Simple DES 알고리즘의 subkey를 구하는 과정]\n" );
 uval n = p10( k );
 // key 를 5 bit 씩 두 갈래로 분할합니다.
 uval l = LEFT5( n );
 uval r = RIGHT5( n );
 uval tmp;
 debugf( "ep(k):\t\t", n );
 // rotate each half left by one
 l = ROT_L5( l, 1 );
 r = ROT_L5( r, 1 );
 tmp = ( l << 5 ) | r;  // join them
 debugf( "ls-1:\t\t", tmp );
 *k1 = p8( tmp );
 debugf( "k1:\t\t", *k1 );
 // rotate each half left by two
 l = ROT_L5( l, 2 );
 r = ROT_L5( r, 2 );
 tmp = ( l << 5 ) | r;  // join them
 debugf( "ls-2:\t\t", tmp );
 *k2 = p8( tmp );
 debugf( "k2:\t\t", *k2 );
 //printf( "\n" );
}

uval F( uval x, uval k ) {
 uval row, col;
 uval rval;
 uval a = ep(x) ^ k;  // expand and add key
 uval l = LEFT4( a );
 uval r = RIGHT4( a );
 // row bits 30, col bits 21
 // lookup left 4 in S0
 row = ( B(3, l ) << 1 ) | B(0, l );
 col = ( B(2, l ) << 1 ) | B(1, l );
 rval = S0[row][col] << 2; // the left 2 bits
 // lookup right 4 in S1
 row = ( B(3, r ) << 1 ) | B(0, r );
 col = ( B(2, r ) << 1 ) | B(1, r );
 rval |= S1[row][col];  // the right 2 bits
 rval = p4(rval);
 return rval;
}

// fk(L,R,SK) = ( L ^ F(R,SK),R )
uval f_k( uval x, uval k ) {
 uval l = LEFT4( x );
 uval r = RIGHT4( x );
 uval y = l ^ F( r, k ); // "add" them, really xor
 //printf( "f_k l xor F = %d\n", y );
 uval rval = ( y << 4 ) | r; // join them
 //printf( "f_k y<<4 = %d\n", rval );
 return rval;
}

uval encrypt( uval in, uval k1, uval k2 ) {
 uval x = ip( in );
 debugf( "after ip\t", x );
 x = f_k( x, k1 );
 debugf( "after f_k\t", x );
 x = sw( x );
 debugf( "after sw\t", x );
 x = f_k( x, k2 );
 debugf( "after f_k\t", x );
 x = ip_inv( x );
 debugf( "after ip_inv\t", x );
 return (unsigned char)x;
}

//using 8bit encrpyt function "encrypt", encrypt string input
void encrypt_des( unsigned short *input,int input_len, unsigned char *output, unsigned short key)
{
	int i;
	uval key1, key2;
	subkeys( key, &key1, &key2 );
	
	for(i=0;i<input_len;i++)
	{
		output[i] = (unsigned char)encrypt( input[i], key1, key2 );
	}
}

//using 8bit dencrpyt function "encrypt", dencrypt string input 
void decrypt_des(unsigned short *input,int input_len, unsigned char *output, unsigned short key)
{
	int i;
	uval key1, key2;
	subkeys( key, &key1, &key2 );
	
	for(i=0;i<input_len;i++)
	{
		output[i] = (unsigned char)encrypt( input[i], key2, key1 );
	}
} 

/*int main( void ) {
 uval input, output;  // 테스트 변수입니다.
 uval key, key1, key2; // 실제 key와 subkey 들입니다.
 uval tmp;

 // uval은 10 bit로 이루어져야 하고, 2 byte 보다 작은지 검사합니다.
 if( sizeof( uval ) < 2 ) {
  printf( "휴스턴, 문제가 생긴 것 같다!\n" );
  printf( "sizeof(uval): %d\n", sizeof( uval ) );
  printf( "더 큰 값을 담을 수 있는 uval 변수의 정의가 필요합니다!\n" );
  exit( 1 );
 }

 // 실험 대상
 input = 0x79; // 'k' = 0x6b 'c' = 0x63, 'y' = 0x79
 key = 0x017;  // 00 0001 0111 -> 학번 6150017에서 오른쪽 3자리는 0 1 7 입니다.
      // 0  1  7




 printf( "key:\t\t" ); p_bits( key );
 subkeys( key, &key1, &key2 );
 printf( "key1:\t\t" ); p_bits( key1 );
 printf( "key2:\t\t" ); p_bits( key2 );

 printf( "평문:\t\t" ); p_bits( input );
 printf( "\n[Simple DES의 암호화 과정]\n" );
// output = encrypt( input, key1, key2 );
encrypt_des(&input, 1, &output, key);
 printf( "암호문:\t\t" ); p_bits( output ); // 0011 0110 = 36

 // 복호화 과정은 subkey 들을 역으로 대입하면 됩니다.
 printf( "\n[Simple DES의 복호화 과정]\n" );
// tmp = encrypt( output, key2, key1 );
	decrypt_des(&output, 1, &tmp, key);
 printf( "복호문:\t\t" ); p_bits( tmp );
printf("%d\n",(unsigned char)tmp);
 // tmp == input 이면, 올바르게 복호화 된 것입니다.
 if( tmp != input ) {
  printf("휴스턴, 문제가 생긴 것 같다!\n");
  printf("입력 값과 복호화 값이 동일하지 않습니다!\n");
  exit(2);
 }

#ifndef DEBUG
 {
  uval x, y, z;
  // 전체 데이터의 범위인 0~255 사이를 검사합니다.
  for( x=0; x<256; ++x ) {
   y = encrypt( x, key1, key2 );
   z = encrypt( y, key2, key1 );
   if( x != z ) {
    printf( "%X에 오류가 있습니다!\n", x );
   }
  }
 }
#endif

 return 0;
}
*/
